package de.renew.gui;

import java.awt.*;
import java.util.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.figures.*;
import CH.ifa.draw.util.*;
import de.renew.simulator.*;
import de.renew.event.*;
import collections.*;

public class TokenBagFigure extends SimpleCompositeFigure
    implements ChildFigure, PlaceEventListener {

    public TokenBagFigure() {
        super();
    }

    private static final int BORDER = 3;
    private static final int SEPARATOR = 1;

    private OffsetLocator fLocator = null;
    private RenewMode mode;
    private Drawing lockDrawing;
    private PlaceInstanceFigure place;
    private PlaceInstance placeInstance;
    private int cacheMult = -1;
    private int cacheTestMult = -1;
    private TextFigure cardFigure;
    private FigureAttributes attributes;
    private int markingAppearance=-1; // force first-time update

    public TokenBagFigure(RenewMode mode,
			  Drawing lockDrawing,
			  PlaceInstanceFigure place,
				  PlaceInstance pi) {
       this(mode,lockDrawing,place,pi,PlaceFigure.CARDINALITY);
    }

    public TokenBagFigure(RenewMode mode,
			  Drawing lockDrawing,
			  PlaceInstanceFigure place,
			  PlaceInstance pi,
			  int markingAppearance) {
       placeInstance = pi;
       this.mode = mode;
       this.lockDrawing = lockDrawing;
       cardFigure = new TextFigure();
       cardFigure.setReadOnly(true);
       cardFigure.setAttribute("FontStyle",new Integer(Font.ITALIC));
       attributes = new FigureAttributes();
       if (place!=null)
	   setParent(place);
       setMarkingAppearance(markingAppearance);
       placeInstance.addPlaceEventListener(this);
    }

    protected Vector basicHandles() {
	//    public Vector handles() {
	Vector handles=super.basicHandles();
	handles.addElement(new NullHandle(this, RelativeLocator.northWest()));
	handles.addElement(new NullHandle(this, RelativeLocator.northEast()));
	handles.addElement(new NullHandle(this, RelativeLocator.southWest()));
	handles.addElement(new NullHandle(this, RelativeLocator.southEast()));
	return handles;
    }

    /*
    public void draw(Graphics g) {
       if (expanded) {
         g.setColor(Color.red);
         Rectangle box=displayBox();
         g.drawRect(box.x,box.y,box.width-1,box.height-1);
       }
       super.draw(g);
    }
    */

    public boolean canBeParent(ParentFigure p) {
       return ( p instanceof PlaceInstanceFigure
	     && ( place == null
               || place == p ));
    }

    private void unsetParent() {
      if (place != null) {
	place.removeChild(this);
	place=null;
      }
    }

    public boolean setParent(ParentFigure figure) {
        if (!canBeParent(figure))
	  return false;
	if (place == figure)
	  return true;
	unsetParent();
        place = (PlaceInstanceFigure)figure;
	if (place == null) {
          fLocator = null;
        } else {
	  place.addChild(this);
          if (fLocator != null)
             fLocator.setBase(place.connectedTextLocator(this));
          else
	     fLocator = new OffsetLocator(place.connectedTextLocator(this));
	  updateLocation();
	}
	return true;
    }

    public ParentFigure parent() { return place; }

    protected void layout() {
       Point partOrigin = displayBox().getLocation();
       partOrigin.translate(BORDER,BORDER);
       Dimension extent = new Dimension(0, 0);

       FigureEnumeration k = figures();
       while (k.hasMoreElements()) {
            Figure f = k.nextFigure();

 	    Dimension partExtent = f.size();
	    Point corner = new Point(
		                        partOrigin.x+partExtent.width,
		                        partOrigin.y+partExtent.height);
            f.basicDisplayBox(partOrigin, corner);

	    extent.width = Math.max(extent.width, partExtent.width);
	    extent.height += partExtent.height+SEPARATOR;
	    partOrigin.y += partExtent.height+SEPARATOR;
        }
        fDisplayBox.width = extent.width + 2*BORDER;
        fDisplayBox.height = extent.height - SEPARATOR + 2*BORDER;
        updateLocation();
	super.layout();
    }

    protected boolean needsLayout() {
        Dimension extent = new Dimension(0, -SEPARATOR + 2*BORDER);

        FigureEnumeration k = figures();
        while (k.hasMoreElements()) {
            Figure f = k.nextFigure();
            extent.width = Math.max(extent.width, f.size().width);
            extent.height += f.size().height + SEPARATOR;
        }
        extent.width += 2*BORDER;
        return !extent.equals(fDisplayBox.getSize());
    }

    public void moveBy(int x, int y) {
        super.moveBy(x, y);
        if (fLocator != null)
            fLocator.moveBy(x, y);
    }

    /**
     * Updates the location relative to the connected figure.
     * The Figure is centered around the located point.
     */
    public void updateLocation() {
        if (fLocator != null) {
            Point p = fLocator.locate(place);
            Rectangle box=displayBox();
            p.x -= box.x + box.width/2;
            p.y -= box.y + box.height/2;

            if (p.x != 0 || p.y != 0) {
                willChange();
                basicMoveBy(p.x, p.y);
                changed();
            }
        }
    }


  /** This event method is called whenever the marking of
   *  the corresponding PlaceInstance changes dramatically
   *  (i.e. by more than one token).
   */
  public void markingChanged(PlaceEvent pe) {
    // I have dropped the plausibilty check that the
    // event must originate from the referenced place instance,
    // because that can lead to racing conditions during the
    // discard phase, when the place instance is nulled.

    Drawing theLockDrawing=lockDrawing;
    if (theLockDrawing==null) {
      // This figure has already been released.
      return;
    }

    try {
      // Lock the drawing during the entire update.
      theLockDrawing.lock();
      if (lockDrawing==null) {
	// A figure was just now concurrently released.
	return;
      }

      if (markingAppearance==PlaceFigure.CARDINALITY) {
	updateCardFigure();
      } else {
	updateTokenFigures();
      }
	
      if (listener() != null)
          listener().figureRequestUpdate(new FigureChangeEvent(this));
      
      // Is it wise to lock the drawing while doing the
      // notifications? Probably yes.
    } finally {
      theLockDrawing.unlock();
    }
  }

  private void updateCardFigure() {
    placeInstance.lock.lock();
    int newMult = placeInstance.getNumberOfTokens();
    int newTestMult = placeInstance.getNumberOfTestedTokens();
    placeInstance.lock.unlock();

    if ((newMult!=cacheMult) || (newTestMult!=cacheTestMult)) {
      cacheMult = newMult;
      cacheTestMult = newTestMult;
      if (newTestMult > 0) {
	if (newMult>0) {
	  cardFigure.setText(newMult+"("+newTestMult+")");
	} else {
	  cardFigure.setText("("+newTestMult+")");
	}
      } else if (newMult > 0) {
	cardFigure.setText(String.valueOf(newMult));
      } else {
	cardFigure.setText(" ");
      }      
    }
  }
  
  private void updateTokenFigures() {
    // Copy marking information, in order to avoid concurrent updates.
    placeInstance.lock.lock();
    
    Vector tempTokens=new Vector();
    Enumeration tokens=placeInstance.getDistinctTestableTokens();
    while (tokens.hasMoreElements()) {
      tempTokens.addElement(tokens.nextElement());
    }
    Object[] newTokens=new Object[tempTokens.size()];
    TimeSet[] newTimeSets=new TimeSet[newTokens.length];
    boolean[] newIsTested=new boolean[newTokens.length];
    tempTokens.copyInto(newTokens);
    for (int i=0;i<newTokens.length;i++) {
      newTimeSets[i]=placeInstance.getFreeTimeSet(newTokens[i]);
      newIsTested[i]=placeInstance.containsTestedToken(newTokens[i]);
    }
    
    placeInstance.lock.unlock();
    
    // Render marking information.
    removeAll();
    willChange();
    
    double currentTime=SearchQueue.getTime();
    boolean expanded=markingAppearance==PlaceFigure.EXPANDED_TOKENS;
    for (int i=0;i<newTokens.length;i++) {
      Object token=newTokens[i];
      TimeSet timeSet=newTimeSets[i];
      boolean isTested=newIsTested[i];
      
      TimeSet extendedTimeSet;
      if (isTested) {
	extendedTimeSet=timeSet.including(currentTime);
      } else {
	extendedTimeSet=timeSet;
      }
      double[] times=extendedTimeSet.asUniqueArray();
      
      for (int j=0;j<times.length;j++) {
	double time=times[j];
	int mult=timeSet.multiplicity(time);
	boolean showTested=isTested && time==currentTime;
	Figure tokenFigure;
	//  	  if (mult==1 && !showTested && time==0) {
	//  	      tokenFigure=mode.getTokenFigure(token,markingAppearance==EXPANDED_TOKENS);
	//  	      if (token instanceof NetInstance) {
	//  		  addClickHandle(new NetInstanceHandle(tokenFigure,null,
	//  						       (NetInstance)token));
	//  	      }
	//  	  } else
	tokenFigure=new MultipleTokenFigure(mode,mult,showTested,time,
					    token,expanded);
	Enumeration attrenum=attributes.definedAttributes();
	while (attrenum.hasMoreElements()) {
	  String attribute=(String)attrenum.nextElement();
	  tokenFigure.setAttribute(attribute,attributes.get(attribute));
	}
	add(tokenFigure);
      }
    }
    
    layout();
    changed();
  }
  
  /** This event occurs when a single token (contained in the
   *  event object) is put into the corresponding PlaceInstance.
   */
  public void tokenAdded(TokenEvent te) {
      markingChanged(new PlaceEvent(te.getPlaceInstance()));
  }

  /** This event occurs when a single token (contained in the
   *  event object) is removed from the corresponding PlaceInstance.
   */
  public void tokenRemoved(TokenEvent te) {
      tokenAdded(te);
  }

  /** This event occurs when a token (contained in the
   *  event object) is tested within the corresponding PlaceInstance.
   */
  public void tokenTested(TokenEvent te) {
      tokenAdded(te);
  }

  /** This event occurs when a token (contained in the
   *  event object) is untested within the corresponding PlaceInstance.
   */
  public void tokenUntested(TokenEvent te) {
      tokenAdded(te);
  }


  public void release() {
    // Later on, I will not even know my place,
    // let alone the drawing. So keep a note.
    Drawing theDrawing=lockDrawing;
    if (theDrawing==null) {
      // This figure is released a second time.
      return;
    }

    try {
      theDrawing.lock();
      if (lockDrawing==null) {
	// This figure has been released concurrently.
      }
     
      if (placeInstance != null) {
	placeInstance.removePlaceEventListener(this);
	// OK: Release reference to place instance so that the place
	// instance can be garbage collected.
	placeInstance = null;
      }
      // OK: Forget about the place.
      unsetParent();
      // Indicate release.
      lockDrawing=null;
      // OK: Forget about all tokens.
      removeAll();
      
      super.release();
    } finally {
      theDrawing.unlock();
    }
  }

  public String toString() {
    if (placeInstance==null) {
      return "Detached TokenBagFigure@"+System.identityHashCode(this);
    }

    StringBuffer output = new StringBuffer().append('{');

    placeInstance.lock.lock();

    Enumeration tokens = placeInstance.getDistinctTestableTokens();
    while (tokens.hasMoreElements()) {
      Object token=tokens.nextElement();
      int mult=placeInstance.getTokenCount(token);
      boolean isTested=placeInstance.containsTestedToken(token);

      output.append(MultipleTokenFigure.getMultString(mult,isTested));
      output.append(token.toString());

      if (tokens.hasMoreElements())
	output.append(", ");
    }

    placeInstance.lock.unlock();

    output.append('}');
    return output.toString();
  }

  public boolean wantSynchronousNotification() {
    return false;
  }

  public void setMarkingAppearance(int markingAppearance) {
    if (this.markingAppearance != markingAppearance
	&& markingAppearance != PlaceFigure.HIGHLIGHT) {
	willChange();
	this.markingAppearance = markingAppearance;
	if (markingAppearance==PlaceFigure.CARDINALITY) {
	    removeAll();
	    add(cardFigure);
	}
	markingChanged(new PlaceEvent(placeInstance));
	layout();
	changed();
    }
  }

  /** Returns whether this figure can be selected.
   */
    //  public boolean isSelectable() { return false; }

  public boolean inspect(DrawingView view, boolean alternate) {
     // undo support can be implemented here
     // (currently disabled for instance drawings)
     boolean done=super.inspect(view,alternate);
     if (alternate && !done) {
	 view.clearSelection();
	 int newMA;
	 if (markingAppearance==PlaceFigure.CARDINALITY) {
	     if (place==null)
		 // in seperate window:
		 newMA=PlaceFigure.EXPANDED_TOKENS;
	     else
		 newMA=PlaceFigure.TOKENS;
	 } else {
	     newMA=PlaceFigure.CARDINALITY;
	 }
	 setAttribute("MarkingAppearance", new Integer(newMA));
	 view.addToSelection(this);
     }
     return done;
  }

  protected PlaceInstance getPlaceInstance() {
     return placeInstance;
  }


  public void update(FigureChangeEvent e) {
      updateLocation();
      super.update(e);
  }

  public void setAttribute(String name, Object value) {
      if (name.equals("MarkingAppearance")) {
	  if (place != null) {
	      place.setAttribute(name,value);
	  }
	  // if we want TokenBagFigures to also
	  // be switchable if there is no PlaceInstanceFigure
	  // assigned, we need:
	  else {
	     setMarkingAppearance(((Integer)value).intValue());
	  }
      } else {
	  if (markingAppearance!=PlaceFigure.CARDINALITY) {
	      attributes.set(name,value);
	  }
	  super.setAttribute(name,value);
      }
  }

  /**
   * Returns the figures with dependencies of superclasses
   * plus the parent figure of this figure.
   **/
  public FigureEnumeration getFiguresWithDependencies() {
    FigureEnumeration superDep = super.getFiguresWithDependencies();
    Vector myDep = new Vector(1);
    myDep.addElement(place);
    return new MergedFigureEnumerator(superDep, 
				      new FigureEnumerator(myDep));
  }

  /**
   * This class is not serializable, although inheriting
   * serializability from its superclass.
   * <p>
   * Reason: Serialization is currently used for Copy&Paste
   * and the Undo machanism only. As such modifications don't
   * make sense for instance drawings, this  figure, which is
   * a part of such drawings, doesn't need to be serializable.
   * </p><p>
   * Second reason: Some fields of this class would lead to
   * a serialization of the compiled net, which is not desirable.
   * </p>
   * @throws java.io.NotSerializableException always.
   **/
  private void writeObject(java.io.ObjectOutputStream out)
      throws java.io.IOException {
    throw new java.io.NotSerializableException(
	      "TokenBagFigure is not serializable!");
  }
    
  /**
   * Deserialization method, behaves like default readObjec
   * method except restoring the connection to the parent of
   * this child figure.
   **/
  private void readObject(java.io.ObjectInputStream in)
      throws java.io.IOException, ClassNotFoundException {
    in.defaultReadObject();
    if (place != null) {
      place.addChild(this);
    } 
    
  }

}

